@@ -3,6 +3,7 @@
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
 
+import os
 import os.path
 import re
 import sys
@@ -132,12 +133,12 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
         with the size of the font.
         """)
 
-    width = Integer(81, config=True,
+    console_width = Integer(81, config=True,
         help="""The width of the console at start time in number
         of characters (will double with `hsplit` paging)
         """)
 
-    height = Integer(25, config=True,
+    console_height = Integer(25, config=True,
         help="""The height of the console at start time in number
         of characters (will double with `vsplit` paging)
         """)
@@ -244,7 +245,7 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
 
         # Initialize protected variables. Some variables contain useful state
         # information for subclasses; they should be considered read-only.
-        self._append_before_prompt_pos = 0
+        self._append_before_prompt_cursor = self._control.textCursor()
         self._ansi_processor = QtAnsiCodeProcessor()
         if self.gui_completion == 'ncurses':
             self._completion_widget = CompletionHtml(self)
@@ -263,7 +264,7 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
         self._kill_ring = QtKillRing(self._control)
         self._prompt = ''
         self._prompt_html = None
-        self._prompt_pos = 0
+        self._prompt_cursor = self._control.textCursor()
         self._prompt_sep = ''
         self._reading = False
         self._reading_callback = None
@@ -426,7 +427,7 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
         # Manually adjust the scrollbars *after* a resize event is dispatched.
         elif etype == QtCore.QEvent.Resize and not self._filter_resize:
             self._filter_resize = True
-            QtGui.qApp.sendEvent(obj, event)
+            QtGui.QApplication.instance().sendEvent(obj, event)
             self._adjust_scrollbars()
             self._filter_resize = False
             return True
@@ -472,12 +473,12 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
         # a fudge factor of one character here.
         # Note 2: QFontMetrics.maxWidth is not used here or anywhere else due
         # to a Qt bug on certain Mac OS systems where it returns 0.
-        width = font_metrics.width(' ') * self.width + margin
+        width = font_metrics.width(' ') * self.console_width + margin
         width += style.pixelMetric(QtGui.QStyle.PM_ScrollBarExtent)
         if self.paging == 'hsplit':
             width = width * 2 + splitwidth
 
-        height = font_metrics.height() * self.height + margin
+        height = font_metrics.height() * self.console_height + margin
         if self.paging == 'vsplit':
             height = height * 2 + splitwidth
 
@@ -780,7 +781,7 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
         if self.font_size:
             font.setPointSize(self.font_size)
         else:
-            font.setPointSize(QtGui.qApp.font().pointSize())
+            font.setPointSize(QtGui.QApplication.instance().font().pointSize())
         font.setStyleHint(QtGui.QFont.TypeWriter)
         self._set_font(font)
 
@@ -888,18 +889,10 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
             if insert != self._insert_plain_text:
                 self._flush_pending_stream()
             cursor.movePosition(QtGui.QTextCursor.End)
-        start_pos = cursor.position()
 
         # Perform the insertion.
         result = insert(cursor, input, *args, **kwargs)
 
-        # Adjust the prompt position if we have inserted before it. This is safe
-        # because buffer truncation is disabled when not executing.
-        if before_prompt and (self._reading or not self._executing):
-            diff = cursor.position() - start_pos
-            self._append_before_prompt_pos += diff
-            self._prompt_pos += diff
-
         return result
 
     def _append_block(self, block_format=None, before_prompt=False):
@@ -1124,23 +1117,9 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
         alt_down = event.modifiers() & QtCore.Qt.AltModifier
         shift_down = event.modifiers() & QtCore.Qt.ShiftModifier
 
-        #------ Special sequences ----------------------------------------------
-
-        if event.matches(QtGui.QKeySequence.Copy):
-            self.copy()
-            intercepted = True
-
-        elif event.matches(QtGui.QKeySequence.Cut):
-            self.cut()
-            intercepted = True
-
-        elif event.matches(QtGui.QKeySequence.Paste):
-            self.paste()
-            intercepted = True
-
         #------ Special modifier logic -----------------------------------------
 
-        elif key in (QtCore.Qt.Key_Return, QtCore.Qt.Key_Enter):
+        if key in (QtCore.Qt.Key_Return, QtCore.Qt.Key_Enter):
             intercepted = True
 
             # Special handling when tab completing in text mode.
@@ -1167,10 +1146,15 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
                         self.execute(interactive = not shift_down)
                     else:
                         # Do this inside an edit block for clean undo/redo.
+                        pos = self._get_input_buffer_cursor_pos()
+                        complete, indent = self._is_complete(
+                            self._get_input_buffer()[:pos], True)
                         cursor.beginEditBlock()
                         cursor.setPosition(position)
                         cursor.insertText('\n')
                         self._insert_continuation_prompt(cursor)
+                        if indent:
+                            cursor.insertText(indent)
                         cursor.endEditBlock()
 
                         # Ensure that the whole input buffer is visible.
@@ -1319,9 +1303,25 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
                 if not self._reading:
                     if self._tab_pressed():
                         # real tab-key, insert four spaces
-                        cursor.insertText(' '*4)
+                        step = 4 - (self._get_cursor().columnNumber() -
+                                    len(self._continuation_prompt)) % 4
+                        cursor.insertText(' '*step)
                     intercepted = True
 
+            elif key == QtCore.Qt.Key_Backtab:
+                cur = self._get_cursor()
+                cur.movePosition(QtGui.QTextCursor.StartOfLine)
+                cur.movePosition(QtGui.QTextCursor.Right,
+                                 QtGui.QTextCursor.MoveAnchor,
+                                 self._get_prompt_cursor().columnNumber())
+                spaces = self._get_leading_spaces()
+                step = spaces % 4
+                cur.movePosition(QtGui.QTextCursor.Right,
+                                 QtGui.QTextCursor.KeepAnchor,
+                                 min(spaces, step if step != 0 else 4))
+                cur.removeSelectedText()
+                intercepted = True
+
             elif key == QtCore.Qt.Key_Left:
 
                 # Move to the previous line
@@ -1358,7 +1358,18 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
                     start_pos = cursor.position()
                     start_pos += len(self._continuation_prompt)
                     cursor.setPosition(position)
+
+                c = self._get_cursor()
+                spaces = self._get_leading_spaces()
+                if (c.position() > start_pos + spaces or
+                        c.columnNumber() == len(self._continuation_prompt)):
+                    start_pos += spaces     # Beginning of text
+
                 if shift_down and self._in_buffer(position):
+                    if c.selectedText():
+                        sel_max = max(c.selectionStart(), c.selectionEnd())
+                        cursor.setPosition(sel_max,
+                                           QtGui.QTextCursor.MoveAnchor)
                     cursor.setPosition(start_pos, QtGui.QTextCursor.KeepAnchor)
                 else:
                     cursor.setPosition(start_pos)
@@ -1407,6 +1418,21 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
                     intercepted = (not self._in_buffer(anchor) or
                                    not self._in_buffer(position))
 
+        #------ Special sequences ----------------------------------------------
+
+        if not intercepted:
+            if event.matches(QtGui.QKeySequence.Copy):
+                self.copy()
+                intercepted = True
+
+            elif event.matches(QtGui.QKeySequence.Cut):
+                self.cut()
+                intercepted = True
+
+            elif event.matches(QtGui.QKeySequence.Paste):
+                self.paste()
+                intercepted = True
+
         # Don't move the cursor if Control/Cmd is pressed to allow copy-paste
         # using the keyboard in any part of the buffer. Also, permit scrolling
         # with Page Up/Down keys. Finally, if we're executing, don't move the
@@ -1493,10 +1519,6 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
         pos = cursor.position()
         self._flush_pending_stream()
         cursor.movePosition(QtGui.QTextCursor.End)
-        diff = cursor.position() - pos
-        if diff > 0:
-            self._prompt_pos += diff
-            self._append_before_prompt_pos += diff
 
     def _flush_pending_stream(self):
         """ Flush out pending text into the widget. """
@@ -1555,6 +1577,11 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
         cursor.movePosition(QtGui.QTextCursor.End)
         return cursor
 
+    def _get_end_pos(self):
+        """ Convenience method that returns the position of the last character.
+        """
+        return self._get_end_cursor().position()
+
     def _get_input_buffer_cursor_column(self):
         """ Returns the column of the cursor in the input buffer, excluding the
             contribution by the prompt, or -1 if there is no such column.
@@ -1639,6 +1666,26 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
                 break
         return ret[::-1]
 
+    def _get_leading_spaces(self):
+        """ Convenience method that returns the number of leading spaces.
+        """
+        cur = self._get_cursor()
+        cur.select(QtGui.QTextCursor.LineUnderCursor)
+        text = cur.selectedText()[len(self._continuation_prompt):]
+        return len(text) - len(text.lstrip())
+
+
+    @property
+    def _prompt_pos(self):
+        """Find the position in the text right after the prompt"""
+        return min(self._prompt_cursor.position() + 1, self._get_end_pos())
+
+    @property
+    def _append_before_prompt_pos(self):
+        """Find the position in the text right before the prompt"""
+        return min(self._append_before_prompt_cursor.position(),
+                   self._get_end_pos())
+
     def _get_prompt_cursor(self):
         """ Convenience method that returns a cursor for the prompt position.
         """
@@ -1678,7 +1725,7 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
             them. (This emulates the behavior of bash, emacs, etc.)
         """
         document = self._control.document()
-        end = self._get_end_cursor().position()
+        end = self._get_end_pos()
         while position < end and \
                   not is_letter_or_number(document.characterAt(position)):
             position += 1
@@ -1753,7 +1800,7 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
 
         if (self._executing and not flush and
                 self._pending_text_flush_interval.isActive() and
-                cursor.position() == self._get_end_cursor().position()):
+                cursor.position() == self._get_end_pos()):
             # Queue the text to insert in case it is being inserted at end
             self._pending_insert_text.append(text)
             if buffer_size > 0:
@@ -1788,6 +1835,10 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
                         cursor.joinPreviousEditBlock()
                         cursor.deletePreviousChar()
 
+                        if os.name == 'nt':
+                            cursor.select(QtGui.QTextCursor.Document)
+                            cursor.removeSelectedText()
+
                     elif act.action == 'carriage-return':
                         cursor.movePosition(
                             cursor.StartOfLine, cursor.KeepAnchor)
@@ -2070,10 +2121,20 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
             If set, a new line will be written before showing the prompt if
             there is not already a newline at the end of the buffer.
         """
-        # Save the current end position to support _append*(before_prompt=True).
         self._flush_pending_stream()
         cursor = self._get_end_cursor()
-        self._append_before_prompt_pos = cursor.position()
+
+        # Save the current position to support _append*(before_prompt=True).
+        # We can't leave the cursor at the end of the document though, because
+        # that would cause any further additions to move the cursor. Therefore,
+        # we move it back one place and move it forward again at the end of
+        # this method. However, we only do this if the cursor isn't already
+        # at the start of the text.
+        if cursor.position() == 0:
+            move_forward = False
+        else:
+            move_forward = True
+            self._append_before_prompt_cursor.setPosition(cursor.position() - 1)
 
         # Insert a preliminary newline, if necessary.
         if newline and cursor.position() > 0:
@@ -2081,7 +2142,6 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
                                 QtGui.QTextCursor.KeepAnchor)
             if cursor.selection().toPlainText() != '\n':
                 self._append_block()
-                self._append_before_prompt_pos += 1
 
         # Write the prompt.
         self._append_plain_text(self._prompt_sep)
@@ -2100,7 +2160,11 @@ class ConsoleWidget(MetaQObjectHasTraits('NewBase', (LoggingConfigurable, superQ
                 self._prompt_html = None
 
         self._flush_pending_stream()
-        self._prompt_pos = self._get_end_cursor().position()
+        self._prompt_cursor.setPosition(self._get_end_pos() - 1)
+
+        if move_forward:
+            self._append_before_prompt_cursor.setPosition(
+                self._append_before_prompt_cursor.position() + 1)
         self._prompt_started()
 
     #------ Signal handlers ----------------------------------------------------
